#!/usr/bin/env python

'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

import scif
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(description="scientific filesystem tools")


    # Global
    
    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument('--quiet', dest="quiet", 
                        help="suppress print output", 
                        default=False, action='store_true')

    parser.add_argument('--writable','-w', dest="writable", 
                        help="for relevant commands, if writable SCIF is needed", 
                        default=False, action='store_true')

    description = 'actions for Scientific Filesystem'
    subparsers = parser.add_subparsers(help='scif actions',
                                       title='actions',
                                       description=description,
                                       dest="command")

    version = subparsers.add_parser("version",
                                    help="show software version")
 

    # Shell and Interactive Terminal (it)

    pyshell = subparsers.add_parser("pyshell",
                                     help="Interactive python shell to scientific filesystem")

    pyshell.add_argument("recipe", nargs='*',
                          help="recipe file or scientific filesystem base", 
                          type=str)


    shell = subparsers.add_parser("shell",
                                  help="shell to interact with scientific filesystem")

    shell.add_argument("app", nargs='?',
                       help="app shell to, defaults to SCIF base if not set.", 
                       type=str)


    preview = subparsers.add_parser("preview",
                                     help="preview changes to a filesytem")

    # if the user provides more than one argument here, they are apps
    preview.add_argument("recipe", nargs="*", 
                         help="recipe file for the filesystem", 
                         type=str)


    # Help
    help = subparsers.add_parser("help",
                                 help="look at help for an app, if it exists.")


    help.add_argument("app", nargs='*',
                       help="app(s) to print help for", 
                       type=str)


    # Install

    install = subparsers.add_parser("install",
                                     help="install a recipe on the filesystem")

    install.add_argument("recipe", nargs="*", 
                         help="recipe file for the filesystem", 
                         type=str)


    # Inspect

    inspect = subparsers.add_parser("inspect",
                                     help="inspect an attribute for a scif installation")

    inspect.add_argument('attributes',
                         nargs='*',
                         help='''attribute to inspect (runscript|r),
                                                      (environment|e),
                                                      (labels|l), or (all|a) (default)''')


    # Run

    run = subparsers.add_parser("run",
                                 help="entrypoint to run a scientific filesystem")


    run.add_argument("cmd", nargs='*',
                     help="app and optional arguments to target for the entry", 
                     type=str)


    # List and dump
    ls = subparsers.add_parser("apps",
                                help="list apps installed")

    ls.add_argument('-l', dest="longlist", 
                     help="show long listing, including paths", 
                     default=False, action='store_true')


    dump = subparsers.add_parser("dump",
                                help="dump recipe")




    # Execute

    execute = subparsers.add_parser("exec",
                                     help="execute a command to a scientific filesystem")


    execute.add_argument("cmd", nargs='*',
                         help="app and command to execute. Eg, exec appname echo $SCIF_APPNAME", 
                         type=str)

    # Not sure if this is needed.
    #write = subparsers.add_parser("write",
    #                               help="write a Dockerfile (default) or Singularity Recipe.")

    #write.add_argument('--write',
    #                   default='docker',
    #                   const='docker',
    #                   nargs='?',
    #                   choices=['docker', 'singularity'],
    #                   help='write a container build specification from the scif recipe (default: %(default)s)')
        
    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''

    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():
    '''main is the entrypoint to managing or interacting with an scif 
       organized filesystem.
    '''

    parser = get_parser()
    subparsers = get_subparsers(parser)

    def help(return_code=0):
        '''print help, including the software version and active client 
           and exit with return code.
        '''
        version = scif.__version__

        print("\nScientific Filesystem [v%s]" %(version))
        parser.print_help()
        sys.exit(return_code)
    
    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()
    try:
        args, unknown = parser.parse_known_args()
    except:
        sys.exit(0)

    # If unknown arguments were provided, pass on to cmd to run
    if args.command in ['run', 'exec'] and len(unknown) > 0:
        args.cmd += unknown

    # if environment logging variable not set, make silent
    if args.debug is False:
        os.environ['SCIF_MESSAGELEVEL'] = 'INFO'
        os.putenv('SCIF_MESSAGELEVEL', 'INFO')

    if args.quiet is True:
        os.environ['SCIF_MESSAGELEVEL'] = 'QUIET'
        os.putenv('SCIF_MESSAGELEVEL', 'QUIET')

    # Show the version and exit
    if args.command == "version":
        print(scif.__version__)
        sys.exit(0)

    # Does the user want a shell?   
    if args.command == "apps": from .list import main
    if args.command == "dump": from .dump import main
    if args.command == "exec": from .execute import main
    if args.command == "help": from .help import main
    if args.command == "install": from .install import main
    if args.command == "preview": from .preview import main
    if args.command == "pyshell": from .pyshell import main
    if args.command == "shell": from .shell import main
    if args.command == "inspect": from .inspect import main
    if args.command == "run": from .run import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args,
             parser=parser,
             subparser=subparsers[args.command])
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)

if __name__ == '__main__':
    main()



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import sys
import pwd
import os


def main(args,parser,subparser):

    from scif.main import ScifRecipe
    apps = args.recipe

    if len(apps) == 0:
        bot.error("You must provide a recipe (.scif) file to install.")
        sys.exit(1)

    recipe = apps.pop(0)

    if not os.path.exists(recipe):
        bot.error("Cannot find recipe file %s" %recipe)
        sys.exit(1)

    if len(apps) == 0:
        apps = None

    client = ScifRecipe(recipe) # writable is True

    # Preview the entire recipe, or the apps chosen
    client.install(apps)



from .terminal import (
    run_command,
    check_install,
    get_installdir,
    get_thumbnail
)
from .fileio import (
    copyfile,
    mkdir_p,
    write_file,
    write_json,
    which,
    read_file,
    read_json
)



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import sys
import os

    

def main(args,parser,subparser):

    from scif.main import ScifRecipe
    client = ScifRecipe(writable=False, quiet=True)
    longlist = args.longlist
    
    result = []
    for app in client.apps():
        config = client.get_appenv(app)

        # Long listing includes a number with path to app
        if longlist is True:
            result.append([app.rjust(10), config['SCIF_APPROOT']]) 
        else:
            result.append(app.rjust(10))

    if len(result) > 0:

        if longlist is True:
            header = "[app]              [root]"
            bot.custom(prefix="SCIF", message=header, color="CYAN")
            bot.table(result)
        else:
            bot.info('\n'.join(result))



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
from scif.defaults import SCIF_BASE


import sys
import os


def parse_input_preferences(recipe, quiet=False):
    '''parse a recipe|app, meaning a list of one or more things that could be

       any of the following:

        Returns
        =======
        parsed: a dictionary lookup with keys for app, recipe, and quiet.
    '''

    # if coming from inspect, will be string
    if not isinstance(recipe,list):
        recipe = [recipe]

    # First case: recipe and app
    if len(recipe) > 1:

        app = recipe[1]
        recipe = recipe[0]    
        quiet = True

    # Second case, no input or recipe|app
    else:

        app = None

        # If no recipe provided, assume loading base

        if len(recipe) == 0:
            recipe = SCIF_BASE

        # Otherwise, we need to figure out if recipe or base
        else:
            recipe = recipe[0]

            # If recipe doesn't exist as a file, assume it's an app
            if not os.path.exists(recipe):
                app = recipe
                recipe = SCIF_BASE        

            # If the recipe provided is a file, no app detection
            else:
               quiet = True

    # Return the parsed data structure
    parsed = {'quiet': quiet,
              'app': app,
              'recipe': recipe }

    return parsed



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

import errno
import os
import re
import shutil
import requests

import codecs
import json
from scif.logger import bot
import sys


############################################################################
## FOLDER OPERATIONS #######################################################
############################################################################


def mkdir_p(path):
    '''mkdir_p attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.error("Error creating path %s, exiting." % path)
            sys.exit(1)


def which(software):
    '''which is a substitute for shutil.which, which is only supported in
       python 3
    
       Parameters
       ==========
       software: the name of the executable to find

    '''
    path = os.getenv('PATH')
    for p in path.split(os.path.pathsep):
        p = os.path.join(p, software)
        if os.path.exists(p) and os.access(p,os.X_OK):
            return p


############################################################################
## FILE OPERATIONS #########################################################
############################################################################

def copyfile(source, destination, force=True):
    '''copy a file from a source to its destination.
    '''
    if os.path.exists(destination) and force is True:
        os.remove(destination)
    shutil.copyfile(source, destination)
    return destination


def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with codecs.open(filename, mode, encoding='utf-8') as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file
    :param json_obj: the dict to print to json
    :param filename: the output file to write to
    :param pretty_print: if True, will use nicer formatting
    '''
    with codecs.open(filename, mode, encoding='utf-8') as filey:
        if print_pretty:
            filey.writelines(
                json.dumps(
                    json_obj,
                    indent=4,
                    separators=(
                        ',',
                        ': '))
                )
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def read_file(filename, mode="r", readlines=True):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with codecs.open(filename, mode, encoding='utf-8') as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
    the data structure as dict.
    '''
    with codecs.open(filename, mode, encoding='utf-8') as filey:
        data = json.load(filey)
    return data



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''


from scif.logger import bot
from subprocess import (
    Popen,
    PIPE,
    STDOUT
)
import os


################################################################################
# Software Versions
################################################################################

def check_install(software, quiet=True):
    '''attempt to run the command for some software, and
       return True if installed. The command line utils will not run 
       without this check.

       Parameters
       ==========
       software: the software to run.
       command: the command to run. If not provided, will use --version
       quiet: if True, don't print verbose output

    '''
    if command is None:
        command = '--version'
    cmd = [software, command]
    try:
        version = run_command(cmd,software)

    except: # FileNotFoundError
        return False
    if version is not None:
        if quiet is False and version['return_code'] == 0:
            version = version['message']
            bot.info("Found %s version %s" % (software.upper(), version))
        return True 
    return False


def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def get_thumbnail():
    '''return the robot.png thumbnail from the database folder.
       if the user has exported a different image, use that instead.
    '''
    from sregistry.defaults import SREGISTRY_THUMBNAIL
    if SREGISTRY_THUMBNAIL is not None:
        if os.path.exists(SREGISTRY_THUMBNAIL):
            return SREGISTRY_THUMBNAIL
    return "%s/database/robot.png" %get_installdir()


def run_command(cmd, sudo=False):
    '''run_command uses subprocess to send a command to the terminal.
    :param cmd: the command to send, should be a list for subprocess
    :param error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.
    :param sudopw: if specified (not None) command will be run asking for sudo
    '''
    if sudo is True:
        cmd = ['sudo'] + cmd

    output = Popen(cmd,stderr=STDOUT,stdout=PIPE)
    t = output.communicate()[0],output.returncode
    output = {'message':t[0],
              'return_code':t[1]}

    return output



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from collections import OrderedDict
from scif.logger import bot
from scif.utils import read_file
from scif.defaults import sections
import os
import re
import sys

def load_filesystem(base, quiet=False):
    '''load a filesystem based on a root path, which is usually /scif

    Parameters
    ==========
    base: base to load.

    Returns
    =======
    config: a parsed recipe configuration for SCIF
    '''
    from scif.defaults import SCIF_APPS

    if os.path.exists(SCIF_APPS):
        apps = os.listdir(SCIF_APPS)
        config = {'apps': {}}
        for app in apps:
            path = '%s/%s/scif/%s.scif' %(SCIF_APPS, app, app)
            if os.path.exists(path):
                recipe = load_recipe(path)
                config['apps'][app] = recipe['apps'][app]

        if len(config['apps']) > 0:
            if quiet is False:
                bot.info('Found configurations for %s scif apps' %len(config['apps']))
                bot.info('\n'.join(list(config['apps'].keys())))
            return config


def load_recipe(path):
    '''load will return a loaded in (user) scif configuration file

    Parameters
    ==========
    path: a path to a deid file

    Returns
    =======
    config: a parsed recipe configuration for SCIF
    '''

    path = os.path.abspath(path)
    if os.path.exists(path):

        # Read in spec, skip empty lines, don't strip remaining
        spec = [x.strip('\n') 
               for x in read_file(path) 
               if not re.match('^\s+$', x.strip('\n'))]

        spec = [x for x in spec if x not in ['', None]]
        config = OrderedDict()
        section = None
        name = None

        while len(spec) > 0:

            # Clean up white trailing/leading space
            line = spec.pop(0)
            stripped = line.strip()

            # Comment
            if stripped.startswith("#"):
                continue

            # A new section?
            elif stripped.startswith('%'):

                # Remove any comments
                line = line.split('#',1)[0].strip()

                # Is there a section name?
                parts = line.split(' ')
                if len(parts) > 1:
                    name = ' '.join(parts[1:])          
                section = re.sub('[%]|(\s+)','',parts[0]).lower()
                config = add_section(config=config,
                                     section=section,
                                     name=name)

            # If we have a section, and are adding it
            elif section is not None:
                spec = [line] + spec
                config = read_section(config=config,
                                      spec=spec,
                                      section=section,
                                      name=name)


    else:
        bot.debug("Cannot find recipe file %s" %path)
    return config


def read_section(config, spec, section, name):
    '''read in a section to a list, and stop when we hit the next section
    '''
    members = []

    # If we need to add other sections, for now all are "apps"
    global_section = 'apps'

    while True:

        if len(spec) == 0:
            break
        next_line = spec[0]                
        if next_line.upper().strip().startswith("%"):
            break
        else:
            new_member = spec.pop(0)
            if not new_member.strip().startswith('#'):
                members.append(new_member)

    # Add the list to the config
    if len(members) > 0:
        if section is not None and name is not None:
            config[global_section][name][section] = members
        else: # section is None, is just global
            config[global_section] = members

    return config


def add_section(config, section, name=None):
    '''add section will add a section (and optionally)
    section name to a config

    Parameters
    ==========
    config: the config (dict) parsed thus far
    section: the section type (e.g., appinstall)
    name: an optional name, added as a level (e.g., google-drive)

    Resulting data structure is:

            config['registry']['apprun']
            config[name][section]

    '''

    if section is None:
        bot.error('You must define a section (e.g. %appenv) before any action.')
        sys.exit(1)

    if section not in sections:
        bot.error("%s is not a valid section." %section)
        sys.exit(1)

    global_section = 'apps'

    # Add the global section, if doesn't exist
    if global_section not in config:
        config[global_section] = OrderedDict()

    if name is not None:        
        if name not in config[global_section]:
            config[global_section][name] = OrderedDict()

        if section not in config[global_section][name]: 
            config[global_section][name][section] = []
            bot.debug("Adding section %s %s" %(name, section))

    return config



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
import sys
import pwd
import os

    

def main(args,parser,subparser):

    from scif.main import ScifRecipe
    cmd = args.cmd

    if len(cmd) < 2:
        bot.warning('You must supply an appname and command to execute.')
        bot.custom(prefix="Example: ", message="scif exec app echo $SCIF_APPNAME")
        sys.exit(1)

    app = cmd.pop(0)
    client = ScifRecipe(quiet=True, writable=args.writable)
    client.execute(app, cmd)



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
from scif.main.helpers import parse_entrypoint
from scif.defaults import SCIF_ENTRYFOLDER, SCIF_SHELL
import sys
import os


def app(self, app):
    '''view a single app, if it exists

    Parameters
    ==========
    app: the name of the app to view
        '''
    if 'apps' in self._config:
        if app in self._config['apps']:
            return self._config['apps'][app]

def apps(self):
    '''get a list of apps to show the user
    '''
    apps = []
    if self._config is not None:
        if "apps" in self._config:
            apps = list(self._config['apps'])
    return apps


def activate(self, app, cmd=None, args=None):
    '''if an app is valid, get it's environment to make it active.
       Update the entrypoint to be relevant to the app runscript.
    
    Parameters
    ==========
    app: the name of the app to activate
    cmd: if defined, the entry point (command) to run. Otherwise uses apprun
    args: additional commands for the apprun

    '''
    if app is None:
        bot.warning('No app selected, will run default %s' %self._entry_point)    
        self.reset()

    elif app in self.apps():
        config = self.get_appenv(app)

        # Make app active
        self._active = app

        # Make sure bin is added to path, and lib to ld_library_path
        self.append_path('PATH', config['SCIF_APPBIN'])
        self.append_path('LD_LIBRARY_PATH', config['SCIF_APPLIB'])

        # Set the runscript, first to cmd provided (exec) then runscript
        if cmd is not None:
            self._entry_point = parse_entrypoint(cmd)
        
        elif 'SCIF_APPRUN' in config:
            if os.path.exists(config['SCIF_APPRUN']):
                self._entry_point = [SCIF_SHELL, config['SCIF_APPRUN']]
                if args is not None:
                    args = parse_entrypoint(args)
                    self._entry_point += args

        # Update the environment for active app (updates ScifRecipe object)
        appenv = self.get_appenv(app, isolated=False, update=True)

        self.load_env(app)  # load environment variables from app itself
        self.export_env()   # export all variables from client.environment

        # Only set entryfolder if user didn't set to something else
        if not SCIF_ENTRYFOLDER:
            self._entry_folder = appenv['SCIF_APPROOT']

    else:        
        bot.warning('%s is not an installed SCIF app' %app)


def deactivate(self, app):
    '''if an app is valid, change the state so the app is no longer active.
       This is currently equivalent to calling reset, but only doing so if the
       app is defined for the SCIF.

    Parameters
    ==========
    app: the name of the app to deactivate
    '''

    if app in self.apps():
        self.reset()

    else:        
        bot.warning('%s is not an installed SCIF app' %app)


def help(self, app):
    '''print the help file for an app, if it exists.

       Parameters
       ==========
       app: the app to export variables for

    '''
    lines = None
    if app in self.apps():
        config = self.get_appenv(app)

        if 'SCIF_APPHELP' in config:
            helpfile = config['SCIF_APPHELP']
            if os.path.exists(helpfile):
                with open(helpfile, 'r') as filey:
                    lines = filey.readlines()
                print('' .join(lines))

    if lines is None:
        bot.info('No help is defined for %s' %app)
    return lines


def reset(self):
    '''reset the SCIF filesystem, meaning that defaults are set, the entrypoint
      if reset, and the environment is reset. Only maintain entry folder set
      by environment, if it was defined.
    '''

    self.set_defaults()
    # Make the active app None
    # self._active = None
    # Update the entry point to use default
    # self._entry_point = parse_entrypoint()
 
    # Reset the environment
    self.update_env(reset=True)

    # set entry folder back to use preference, if original is not None
    if not SCIF_ENTRYFOLDER:
        self._entry_folder = SCIF_ENTRYFOLDER        


def inspect(self, app, attributes=None):
    '''inspect an app based on a list of attributes to inspect.

    Parameters
    ==========
    app: the name of the app to inspect
    attributes: a list of attributes to return
    '''
    result = {}
    if app not in self.apps():
        return result

    if attributes is None:
        attributes = ['all']

    lookup = self.app(app)

    if 'a' in attributes or 'all' in attributes:
        return lookup

    if 'h' in attributes or 'help' in attributes and 'apphelp' in lookup:
        result['apphelp'] = lookup['apphelp']
    if 'f' in attributes or 'files' in attributes and 'appfiles' in lookup:
        result['appfiles'] = lookup['appfiles']
    if 'r' in attributes or 'runscript' in attributes and 'apprun' in lookup:
        result['apprun'] = lookup['apprun']
    if 'l' in attributes or 'labels' in attributes and 'applabels' in lookup:
        result['applabels'] = lookup['applabels']
    if 'e' in attributes or 'environment' in attributes and 'appenv' in lookup:
        result['appenv'] = lookup['appenv']
    if 'i' in attributes or 'install' in attributes and 'appinstall' in lookup:
        result['appinstall'] = lookup['appinstall']

    return result



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.logger import bot
from scif.utils import mkdir_p
from scif.main.helpers import parse_entrypoint
import sys
import os


def set_base(self, base='/', writable=True):
    ''' set the base (the root where to create /scif) and determine if
        it is writable

        Parameters
        ==========
        base: the full path to the root folder to create /scif
    '''
    # The user is likely to give path to scif (should still work)
    base = base.strip('scif')

    if not os.path.exists(base):
        bot.error('%s does not exist!' %base)
        sys.exit(1)

    base = "/%s" %os.path.abspath(base).strip('/')
    self._base = os.path.join(base,'scif')
    self.path_apps = '%s/apps' %self._base
    self.path_data = '%s/data' %self._base

    # Update the environment
    self.add_env('SCIF_DATA', self.path_data)
    self.add_env('SCIF_APPS', self.path_apps)

    # Check if it's writable
    if writable is True:
        if not os.access(base, os.W_OK):
            bot.error('%s is not writable.' %base)
            sys.exit(1)



def set_defaults(self, entry_point=None):
    '''set the defaults for the SCIF state at start up. 
       Usually this means parsing them from the environment. We hold these
       with the object so we can easily maintain and change state.
    '''

    # The entrypoint is the runscript (or /bin/bash default)
    self._entry_point = parse_entrypoint(entry_point)   
                                          
    # Set the default entry folder, changed to app if in context of app
    from scif.defaults import SCIF_ENTRYFOLDER
    self._entry_folder = SCIF_ENTRYFOLDER       # Defaults to None

    # The active app, coinciding with the entry_point
    self._active = None


def install_base(self):
    ''' create basic scif structure at the base for apps and metadata

        Parameters
        ==========
        base: the full path to the root folder to create /scif
    '''
    if not hasattr(self,'_base'):
        bot.error('Please set the base before installing to it.')
        sys.exit(1)

    bot.info('Installing base at %s' %self._base)  

    mkdir_p(self.path_apps)
    mkdir_p(self.path_data)



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

__version__ = "0.0.65"
AUTHOR = 'Vanessa Sochat'
AUTHOR_EMAIL = 'vsochat@stanford.edu'
NAME = 'scif'
PACKAGE_URL = "http://www.github.com/containers-ftw/scif-cli"
KEYWORDS = 'the scientific filesystem'
DESCRIPTION = "a filesystem organization for scientific software and metadata"
LICENSE = "LICENSE"

INSTALL_REQUIRES = (
    ('python-dateutil', {'exact_version': "2.5.3"}),
    ('requests', {'exact_version': '2.18.4'}),
    ('requests-toolbelt', {'exact_version': '0.8.0'}),
    ('retrying', {'exact_version': '1.3.3'}),
    ('six', {'min_version': '1.7.0'}),
    ('pygments', {'min_version': '2.1.3'})
)

# Submodule Requirements

INSTALL_REQUIRES_ALL = (INSTALL_REQUIRES)



'''

Copyright (C) 2017 The Board of Trustees of the Leland Stanford Junior
University.
Copyright (C) 2016-2017 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from scif.client.utils import parse_input_preferences
from scif.logger import bot
import sys
import os


def main(args,parser,subparser):

    from scif.defaults import SCIF_SHELL as shell

    parsed = parse_input_preferences(args.recipe, 
                                     quiet=args.quiet)

    recipe = parsed['recipe']
    quiet = parsed['quiet']
    app = parsed['app']

    # The client will either load the recipe or the base

    lookup = { 'ipython': ipython,
               'python': python,
               'bpython': bpython }

    shells = ['ipython', 'python', 'bpython']

    # If the user asked for a specific shell via environment
    if shell is not None:
        shell = shell.lower()
        if shell in lookup:
            try:    
                return lookup[shell](recipe=recipe, 
                                     app=app, 
                                     quiet=quiet,
                                     writable=args.writable)
            except ImportError:
                pass

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](recipe=recipe, 
                                 app=app, 
                                 quiet=quiet,
                                 writable=args.writable)
        except ImportError:
            pass
    

def ipython(recipe, app=None, quiet=False, writable=True):
    '''embed the client with loaded recipe into an ipython session
    '''
    from scif.main import ScifRecipe
    from IPython import embed
    client = ScifRecipe(recipe, quiet=quiet, writable=writable)
    client.speak()
    if app is not None:
        client.activate(app)
    embed()

def bpython(recipe, app=None, quiet=False, writable=True):
    from scif.main import ScifRecipe
    import bpython
    client = ScifRecipe(recipe, quiet=quiet, writable=writable)
    client.speak()
    if app is not None:
        client.activate(app)
    bpython.embed(locals_={'client': client})

def python(recipe, app=None, quiet=False, writable=True):
    from scif.main import ScifRecipe
    import code
    client = ScifRecipe(recipe, quiet=quiet, writable=writable)
    client.speak()
    if app is not None:
        client.activate(app)
    code.interact(local={"client":client})
